import json
from collections import Counter
from dataclasses import dataclass, field
from datetime import datetime
from enum import auto
from typing import Dict

from ftfy import fix_encoding
from scraptrawler.utils import DecklistFormatter, ExtendedEnum, get_logger
from strenum import StrEnum


class InThe(ExtendedEnum, StrEnum):
    """
    The location of a card in a deck.

    Attributes
    ----------
    MAIN: Main deck
    SIDE: Sideboard
    """

    MAIN = auto()
    SIDE = auto()


class Format(ExtendedEnum, StrEnum):
    """
    The format a deck was played in.

    Attributes
    ----------
    VINTAGE: Vintage
    LEGACY: Legacy
    PAUPER: Pauper
    MODERN: Modern
    PIONEER: Pioneer
    STANDARD: Standard
    """

    VINTAGE = "Vintage"
    LEGACY = "Legacy"
    PAUPER = "Pauper"
    MODERN = "Modern"
    PIONEER = "Pioneer"
    STANDARD = "Standard"


class Deck:
    """
    A class to represent a deck of Magic: the Gathering cards.

    Attributes
    ----------
    archetype : str
        The archetype of this deck.
    format : Deck.Format
        The format of the event this deck was played in.
    date_played : datetime
        The date this deck was played.
    main : dict[str:int]
        The main deck (generally 60 cards).
    side : dict[str:int]
        The sideboard (generally 15 cards).

    Methods
    -------
    add_card(card: str, quantity: int, in_the: InThe):
        Adds a given quantity of a given card to this Deck.
    add_cards(cards: Dict[str, int], in_the: InThe):
        Adds the given cards to this Deck.
    as_cards():
        Returns all the cards in this Deck as a dict[str:int]
    count_main():
        Counts the cards in the main deck.
    count_side():
        Counts the cards in the sideboard.
    count_all():
        Counts all of the cards in this Deck.
    to_json():
        Converts this Deck to JSON.
    to_decklist(DecklistFormat):
        Exports the Deck as a str with the given DecklistFormat.
    """

    ## Class Attributes
    # Set up logger
    log_filename = "deck.log"
    logger = get_logger(log_filename, "deck")

    # Constants
    DEFAULT_ARCHETYPE = ""

    def __init__(
        self,
        archetype: str = DEFAULT_ARCHETYPE,
        format: Format = None,
        date_played: datetime = datetime.today(),
        main: Counter = Counter(),
        side: Counter = Counter(),
    ):
        self.archetype = archetype
        self.date_played = date_played
        self.format = format

        # Deep copy of main and side
        self.main, self.side = Counter(), Counter()
        self.main.update(main)
        self.side.update(side)

    def __eq__(self, other):
        return (
            self.archetype == other.archetype
            and self.date_played == other.date_played
            and self.format == other.format
            and self.main == other.main
            and self.side == other.side
        )

    def add_cards(self, cards: Dict[str, int], in_the: InThe = InThe.MAIN) -> None:
        """
        Adds the given cards to this Deck.

        Parameters:
            cards Dict[str:int]: The cards to add.
            in_the (InThe): Where to add the cards (main, side, etc)
        """
        for c, q in cards.items():
            self.add_card(card=c, quantity=q, in_the=in_the)

    def add_card(self, card: str, quantity: int = 1, in_the: InThe = InThe.MAIN) -> None:
        """
        Adds a given quantity of a given card to this Deck.

        Parameters:
            card (str): The card to add.
            quantity (int): The number of copies of the card to be added.
            in_the (InThe): Where to add the card (main, side, etc)
        """
        card = fix_encoding(card)  # Fixes UTF-8 encoding issues like "3 LÃ³rien Revealed" -> "3 Lórien Revealed"
        match in_the:
            case InThe.MAIN:
                self.main.update({card: quantity})
                self.logger.debug(f"{self.archetype} - Added {quantity} copies of {card} to the main deck.")
            case InThe.SIDE:
                self.side.update({card: quantity})
                self.logger.debug(f"{self.archetype} - Added {quantity} copies of {card} to the sideboard.")
            case _:
                self.logger.error(
                    f"{self.archetype} - Unable to add {quantity} copies of {card} to the deck. 'in' must be one of {InThe.list()}"
                )

    def as_cards(self) -> Counter:
        """
        Returns a Counter containing all the cards in this deck. Combines main and side.

        Returns:
            cards (dict[str: int]): A dict with card names and quantities from main and side combined.
        """
        cards = Counter()
        cards.update(self.main)
        cards.update(self.side)
        return cards

    def count_main(self) -> int:
        """
        Returns:
            int: The number of cards in this Deck's main.
        """
        return len(self.main)

    def count_side(self) -> int:
        """
        Returns:
            int: The number of cards in this Deck's side.
        """
        return len(self.side)

    def count_all(self) -> int:
        """
        Returns:
            int: The number of cards in this Deck.
        """
        return self.count_main() + self.count_side()

    def to_json(self) -> str:
        """
        Returns:
            dict: A JSON object with the data stored in this Deck.
        """
        return json.dumps(self.__dict__, indent=4, ensure_ascii=False, default=str)

    def to_decklist(self, decklist_formatter: DecklistFormatter = None) -> str:
        """
        Exports this Deck as a str with the given DecklistFormatter.

        Parameters:
            export_format (DecklistFormatter): The format of the exported Deck.

        Returns:
            decklist (str): A string containing the names and quantities of the cards in this Deck.
        """

        match decklist_formatter:
            case DecklistFormatter.ARENA:
                sb_prefix = "Sideboard\n"
                # TODO: filter out cards that are not on Arena. Log a WARNING with those cards.
                self.logger.debug(f"{self.archetype} - Exporting for Arena.")
            case DecklistFormatter.MTGO:
                sb_prefix = "SIDEBOARD:\n"
                # TODO: filter out cards that are not on MTGO. Log a WARNING with those cards.
                self.logger.debug(f"{self.archetype} - Exporting for MTGO.")
            case _:
                sb_prefix = ""  # Default
                self.logger.warning(
                    f"""{self.archetype} - Unable to export with the given format: {decklist_formatter}. """
                    f"""'export_format' must be one of {DecklistFormatter.list()}. Using default format."""
                )
        sb_prefix = "\n\n" + sb_prefix

        # Build the decklist string
        main = "\n".join([f"{quantity} {card_name}" for card_name, quantity in self.main.items()])
        side = sb_prefix + "\n".join([f"{quantity} {card_name}" for card_name, quantity in self.side.items()])
        decklist = f"{main}{side if self.count_side() > 0 else ''}"
        return decklist



import logging
from enum import IntEnum
from typing import Counter, List

import requests
from scraptrawler.deck import Deck, InThe
from scraptrawler.utils import ExtendedEnum, get_logger
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from webdriver_manager.chrome import ChromeDriverManager


class LoadTime(ExtendedEnum, IntEnum):
    """
    A class used for standardized load times to be used during WebDriverWait.

    Attributes
    ----------
    STD (5) : Standard load time for interacting with normal DOM elements.
    LONG (10) : Long load time while waiting for AJAX or other non-standard elements.
    """

    STD = 5
    LONG = 10


class BaseScraper:
    """
    A class to represent a web scraper used to get data about
    Magic: the Gathering decklists and results.

    Attributes
    ----------
    driver : webdriver.Remote
        Controls a browser by sending commands to a remote server.

    Methods
    -------
    get_decks_from_event_url(url: str, top8: bool):
        Returns a list of Decks from the given event URL.
    get_deck_from_url(url: str):
        Returns a Deck from the given deck page URL.
    """

    ## Class Attributes
    log_filename = "scraper.log"

    def __init__(
        self,
        driver: webdriver.Remote = None,
        logger: logging.Logger = get_logger(log_filename, "BaseScraper"),
    ):
        self.driver = driver
        self.logger = logger
        self.__post_init__()

    def __post_init__(self):
        if self.driver is None:
            ## Setup Chrome options
            chrome_options = Options()
            chrome_options.add_argument("--headless=new")  # Ensure GUI is off

            # Setup Chrome service
            # NOTE: ChromeDriverManager(driver_version="") refers to the local google-chrome --version
            # NOTE: There is a bug if you have a Chrome version that is too new with no matching chromedriver
            chrome_service = Service(ChromeDriverManager().install())

            # Choose Chrome Browser
            self.driver = webdriver.Chrome(service=chrome_service, options=chrome_options)
            self.driver.maximize_window()

    ### Public Functions

    def get_decks_from_event_url(self, url: str) -> List[Deck]:
        """
        Returns a list of Decks from the given event URL.

        Parameters:
            url (str): The URL to an event.
            top8 (bool): Return only the top 8 Decks, default = False.

        Returns:
            List[Deck]: A list of Decks from the given event URL.
        """
        raise (NotImplementedError)

    def get_deck_from_url(self, url: str) -> Deck:
        """
        Returns a Deck from the given URL.

        Parameters:
            url (str): The URL to a deck.

        Returns:
            deck (Deck): A dict of the cards in the given deck.
        """
        raise (NotImplementedError)

    ### Protected Functions

    def _load(self, url: str) -> None:
        """
        Tries to load a web page in the current browser session.

        Parameters:
            url (str): The URL to a web page (generally an event or decklist page).
        """
        try:
            r = requests.get(url)
            self.logger.debug(f"URL status - URL: {url}, Status: {r.status_code}")
        except requests.exceptions.RequestException as e:
            self.logger.error(e)
            raise SystemExit(e)
        else:
            self.driver.get(url)

    def _sanitized_title(self) -> str:
        """
        Provides a plain-text version of the current web page title.
        Use _load(url) before executing this function.

        Returns:
            title (str): A sanitized title.
        """
        raise (NotImplementedError)

    def _decks_from_elements(self, main_elements: List[WebElement], side_elements: List[WebElement]) -> List[Deck]:
        """
        Returns a list of Decks from the given WebElements.

        Parameters:
            main_elements (List[WebElement]): The WebElements containing main deck cards.
            side_elements (List[WebElement]): The WebElements containing sideboard cards.

        Returns:
            decks (List[Deck]): A list of decks.
        """
        raise (NotImplementedError)

    def _deck_from_parts(self, main_element: WebElement, side_element: WebElement) -> Deck:
        """
        Returns a Deck with the given main deck and sideboard WebElements.

        Parameters:
            main_element (WebElement): A WebElement containing main deck cards.
            side_element (WebElement): A WebElement containing sideboard cards.

        Returns:
            deck (Deck): A Deck.
        """
        raise (NotImplementedError)

    def _cards_from_element(self, deck_element: WebElement) -> List[Counter]:
        """
        Returns a List of cards in the given deck element

        Parameters:
            deck_element (WebElement): A WebElement containing cards.

        Returns:
            cards (List[Counter[str]]): A list of cards.
        """
        raise (NotImplementedError)



import importlib.metadata
import logging
import os.path
from enum import Enum, auto
from sys import stdout

from strenum import StrEnum


## Constants
class Constants:
    MELEE_DATE_FORMAT = "%m/%d/%Y"
    MTGO_DATE_FORMAT = "%B %d, %Y"
    VERSION = importlib.metadata.version("scraptrawler")
    USER_AGENT = f"scraptrawler {VERSION}"
    UUID_PATTERN = "[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}"


def get_logger(
    filename: str,
    logger_name: str,
    file_logging_level: int = logging.DEBUG,
    console_logging_level: int = logging.WARNING,
    formatter: logging.Formatter = logging.Formatter("%(asctime)s - %(name)s:%(levelname)s - %(message)s"),
):
    """
    Helper function to get a new logger.

    Parameters:
        filename (str): Filename of the log file.
        logger_name (str): The logger's name.
        file_logging_level (int): Logging level for the log file.
        console_logging_level (int): Logging level for stdout.
        formatter (logging.Formatter): The Formatter to be used in messages generated by this logger.

    Returns:
        logger (logging.Logger): A new logger.
    """

    # Create directory if not exists
    filepath = os.path.join("logs", filename)
    os.makedirs(os.path.dirname(filepath), exist_ok=True)

    # Handlers
    fh = logging.FileHandler(filepath, mode="a", encoding="UTF-8", delay=False)
    fh.setLevel(file_logging_level)
    ch = logging.StreamHandler(stdout)
    ch.setLevel(console_logging_level)

    # Formatting
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)

    # Create the logger
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(fh)
    logger.addHandler(ch)

    return logger


class ExtendedEnum(Enum):
    """
    An extension of the classic Python Enum to support additional behaviors.

    Methods
    -------
    list():
        Returns a list of the attributes of this Enum.
    get(key):
        Returns the value if it has a matching key. Otherwise, None.
    """

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))

    @classmethod
    def get(cls, key):
        key = str(key).strip().lower()
        for v in cls.list():
            if key == str(v).lower():
                return cls(v)


class DecklistFormatter(ExtendedEnum, StrEnum):
    """
    A method of formatting a decklist for external systems.

    Attributes
    ----------
    ARENA: Magic: the Gathering Arena
    MTGO: Magic: the Gathering Online
    """

    ARENA = auto()
    MTGO = auto()



from collections import namedtuple
from datetime import datetime
from typing import Counter, List

from scraptrawler.deck import Deck, Format, InThe
from scraptrawler.scraper.base import BaseScraper, LoadTime
from scraptrawler.utils import Constants, ExtendedEnum, get_logger
from selenium import webdriver
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from strenum import StrEnum


class MeleeXPath(ExtendedEnum, StrEnum):
    """
    A class used for storing XPath related to MTG Melee and its DOM elements.

    Deck Parts
    ----------
    MAIN : Main deck elements.
    SIDE : Sideboard element.
    QTY : Card quantity elements.
    CARD : Card name elements.
    QTY_DATA : The attribute containing the relevant quantity data.
    CARD_DATA : The attribute containing the relevant card name data.

    Page Info
    ----------
    TOURNAMENT_HEADLINE : Headline at the top of a tournament page. Has date, registration, and other information.

    Standings Table
    ----------
    DECK_URL : Deck URL elements.
    PROCESSING_LBL : Label that appears when the tournament standings table is loading.

    Pagination
    ----------
    PAGINATE_BTN : The page # buttons at the bottom of the tournament standings table.
    PER_PAGE_MAX_OPT : The maximum per-page option.
    PER_PAGE_SELECT : The select tag that determines how many deck URLs are on each page of the tournament standings
    table.
    """

    # Deck parts
    DATE_PLAYED = "//*[local-name() = 'svg'][@data-icon='calendar-days']/parent::div"
    DECK_FORMAT = "//*[local-name() = 'svg'][@data-icon='fort-awesome']/parent::div"
    MAIN = (
        "//tr//td[@class='decklist-builder-section-label-cell']"
        "[not(text() = 'Sideboard')][not(text() = 'Companion')]"
        "/ancestor::table"
    )
    SIDE = "//tr//td[@class='decklist-builder-section-label-cell'][text() = 'Sideboard']/ancestor::table"
    QTY = ".//td[@class='decklist-builder-card-quantity-cell']"
    CARD = ".//a[@data-type='card']"
    QTY_DATA = "innerHTML"
    CARD_DATA = "data-name"

    # Page info
    ACCEPT_COOKIES_BTN = "//div[@class='d-flex flex-row-reverse']//input"
    TOURNAMENT_HEADLINE = "//p[@id='tournament-headline-registration']"

    # Standings table
    DECK_URL = "//table[@id='tournament-standings-table']//a[@data-type='decklist']"
    PROCESSING_LBL = "//div[@id='tournament-standings-table_processing']"

    # Pagination
    PAGINATE_BTN = "//span//a[contains(@class, 'paginate_button')][@aria-controls='tournament-standings-table']"
    PER_PAGE_MAX_OPT = "//select[@aria-controls='tournament-standings-table']//option[@value='500']"
    PER_PAGE_SELECT = "//select[@aria-controls='tournament-standings-table']"


class MeleeScraper(BaseScraper):
    """
    A class to represent a type of Scraper used to get data about Magic: the Gathering decklists and results from
    [MTG Melee](https://melee.gg).

    Attributes
    ----------
    driver : webdriver.Remote
        Controls a browser by sending commands to a remote server.

    Methods
    -------
    get_event_slots_from_event_url(url: str)
        Returns a tuple of (slots_remaining, total_slots)
    get_decks_from_event_url(url: str, top8: bool):
        Returns a list of Decks from the given event URL.
    get_deck_from_url(url: str):
        Returns a Deck from the given deck page URL.
    """

    def __init__(self, driver: webdriver.Remote = None):
        self.logger = get_logger(self.log_filename, "MeleeScraper")
        super().__init__(driver=driver, logger=self.logger)

    ### Melee Public Functions

    def get_slots_from_event_url(self, url: str) -> namedtuple("EventSlots", ["total", "enrolled", "available"]):
        """
        Returns a namedtuple, EventSlots with attributes total, enrolled, and available.

        Parameters:
            url (str): The URL to an event.

        Returns:
            namedtuple(int, int, int): EventSlots(total, enrolled, available)
        """

        # Get page
        self._load(url)
        event_name = self._sanitized_title()
        self.logger.info(f"{url} - Processing event - {event_name}.")
        self.__accept_cookies()

        # Get tournament headline
        tournament_headline = (
            WebDriverWait(self.driver, LoadTime.STD)
            .until(EC.presence_of_element_located((By.XPATH, MeleeXPath.TOURNAMENT_HEADLINE)))
            .text
        )
        enrollment_info = tournament_headline.split("|")[-1]
        enrollment_nums = [int(i) for i in enrollment_info.split() if i.isdigit()]

        # Calculate event slot info
        EventSlots = namedtuple("EventSlots", ["total", "enrolled", "available"])
        event_slots = EventSlots(
            total=enrollment_nums[1],
            enrolled=enrollment_nums[0],
            available=enrollment_nums[1] - enrollment_nums[0],
        )
        return event_slots

    ### Public Function Overrides

    def get_decks_from_event_url(self, url: str, top8: bool = False) -> List[Deck]:
        # Get page
        self._load(url)
        event_name = self._sanitized_title()
        self.logger.info(f"{url} - Processing event - {event_name}.")
        self.__accept_cookies()

        if not top8:
            # Update the decklists per page to 500
            self.__set_per_page_max()
            self.__wait_for_standings()

        # Count the # of pages to scrape
        paginate_btns = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_all_elements_located((By.XPATH, MeleeXPath.PAGINATE_BTN))
        )
        page_count = int(paginate_btns[-1].get_attribute("textContent") if len(paginate_btns) > 1 else 1)

        # Process each page for this event
        deck_urls = []
        for p in range(1, page_count + 1):
            self.logger.debug(f"Processing page {p}/{page_count}.")
            deck_urls.extend(self.__get_deck_urls_from_pagination())

            # Top 8 only?
            if top8:
                break

            # Click to the next page
            next_page_btn = self.driver.find_element(By.ID, "tournament-standings-table_next")
            self.driver.execute_script("$(arguments[0]).trigger('click')", next_page_btn)
            self.__wait_for_standings()

        # TODO: This is a O(2n) way of doing this. If we could open a new window in get_deck_from_url, we could move
        # this to the above loop and do it in O(n). This operation is slow as it is.
        if top8:
            deck_urls = deck_urls[:8]
        decks = []
        for i in range(0, len(deck_urls)):
            print(f"Processing deck {i+1}/{len(deck_urls)}...", end="\r")
            decks.append(self.get_deck_from_url(deck_urls[i]))
        print()  # reset \r endline

        return decks

    def get_deck_from_url(self, url: str) -> Deck:
        # Get page
        self._load(url)
        deck_archetype = self._sanitized_title() or Deck.DEFAULT_ARCHETYPE
        self.logger.info(f"{url} - Processing deck - {deck_archetype}.")
        self.__accept_cookies()

        ## Get Deck parts
        # Format
        deck_format_element = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_element_located((By.XPATH, MeleeXPath.DECK_FORMAT))
        )
        deck_format = Format.get(deck_format_element.text)

        # Date
        date_played_element = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_element_located((By.XPATH, MeleeXPath.DATE_PLAYED))
        )
        date_played = datetime.strptime(date_played_element.text.strip(), Constants.MELEE_DATE_FORMAT)

        # Main and side
        main_elements = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_all_elements_located((By.XPATH, MeleeXPath.MAIN))
        )
        side_element = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_element_located((By.XPATH, MeleeXPath.SIDE))
        )

        # Create Deck
        deck = self._deck_from_parts(
            main_elements=main_elements,
            side_element=side_element,
        )
        deck.archetype = deck_archetype
        deck.date_played = date_played
        deck.format = deck_format

        return deck

    ### Protected Function Overrides

    def _sanitized_title(self) -> str:
        # NOTE: Melee decks and events are sanitized the same way.
        try:
            return self.driver.title.replace(" | Melee", "")
        except Exception as e:
            # raise(e)
            self.logger.warning(f"{self.driver.current_url} - Unable to sanitize title. Proceeding with defaults.")
            return ""

    def _deck_from_parts(self, main_elements: List[WebElement], side_element: WebElement) -> Deck:
        """
        Returns a Deck with the given main deck sections and sideboard WebElements.

        Parameters:
            main_elements (List[WebElement]): A list of WebElements containing main deck cards. Melee has type-based
            subsections.
            side_element (WebElement): A WebElement containing sideboard cards.

        Returns:
            deck (Deck): A Deck.
        """

        # Get the cards
        main_cards = Counter()
        for me in main_elements:
            main_cards.update(self._cards_from_element(me))
        side_cards = self._cards_from_element(side_element)

        # Create a Deck
        deck = Deck()
        deck.add_cards(cards=main_cards, in_the=InThe.MAIN)
        deck.add_cards(cards=side_cards, in_the=InThe.SIDE)

        return deck

    def _cards_from_element(self, deck_element: WebElement) -> List[Counter]:
        ## Get cards and quantities from the page
        quantities = [
            int(qe.get_attribute(MeleeXPath.QTY_DATA))  # WebElement containing quantity
            for qe in WebDriverWait(deck_element, LoadTime.STD).until(
                EC.presence_of_all_elements_located((By.XPATH, MeleeXPath.QTY))
            )
        ]
        card_names = [
            ce.get_attribute(MeleeXPath.CARD_DATA)  # WebElement containing card name
            for ce in WebDriverWait(deck_element, LoadTime.STD).until(
                EC.presence_of_all_elements_located((By.XPATH, MeleeXPath.CARD))
            )
        ]
        cards = dict(zip(card_names, quantities))

        return cards

    ### Melee Private Functions

    def __get_deck_urls_from_pagination(self) -> List[str]:
        """
        Returns a list of deck page URLs from the current WebDriver page.

        Returns:
            List[str]: A list of deck page URLs from the current WebDriver page.
        """

        # Grabs the URL for each decklist on the current page
        atags = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_all_elements_located((By.XPATH, MeleeXPath.DECK_URL))
        )
        deck_urls = [element.get_attribute("href") for element in atags]

        return deck_urls

    def __set_per_page_max(self) -> None:
        """
        Sets the decks-per-page to the maximum value. This will reduce the amount of pagination required to scrape all
        of the decklists from an event page.
        """

        # Find the option tag, set it to selected=true
        per_page_max_opt = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_element_located((By.XPATH, MeleeXPath.PER_PAGE_MAX_OPT))
        )
        self.driver.execute_script("arguments[0].setAttribute('selected', arguments[1])", per_page_max_opt, True)

        # Find the select tag, force it to update
        per_page_select = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_element_located((By.XPATH, MeleeXPath.PER_PAGE_SELECT))
        )
        self.driver.execute_script("$(arguments[0]).trigger('change')", per_page_select)

    def __wait_for_standings(self) -> None:
        """
        Waits for the Tournament Standings table to refresh before continuing to scrape decks.
        """

        # Wait until the tournament standings table is updated
        WebDriverWait(self.driver, LoadTime.LONG).until(
            EC.text_to_be_present_in_element_attribute((By.XPATH, MeleeXPath.PROCESSING_LBL), "style", "display: none")
        )

    def __accept_cookies(self) -> None:
        """
        Clicks the accept button for the request for cookies.
        """
        try:
            WebDriverWait(self.driver, LoadTime.STD).until(
                EC.presence_of_element_located((By.XPATH, MeleeXPath.ACCEPT_COOKIES_BTN))
            ).click()
            self.logger.info(f"Cookies accepted - {self.driver.current_url}")
        except TimeoutException as e:
            self.logger.info("Accept cookies button not found, skipping")



import re
from urllib.parse import urlparse

import requests
from bs4 import BeautifulSoup
from scooze.card import Card
from scooze.deck import Deck, DeckPart
from scraptrawler.utils import Constants

# region Helper Functions


def get_html_document(url: str) -> str:
    """
    GET request for an HTML document from the given URL.
    """

    headers = {"User-Agent": Constants.USER_AGENT}
    response = requests.get(url=url, headers=headers)

    return response.text


def get_apex_from_netloc(netloc: str) -> str:
    """
    Returns the apex domain from a given ParseResult.netloc
    """

    return netloc.split(".")[-2]


def deckpart_from_lines(lines: str) -> DeckPart:
    """
    TODO: docstring explaining how this takes N lines in the form <# card_name> and returns a DeckPart
    TODO: this will be slightly different between deck sites, so maybe add an enum for System to modify behavior
          for MDFCs, only Moxfield has the " // second_face_name"
    """

    part = DeckPart()

    for line in lines:
        quantity, card_name = line.split(" ", maxsplit=1)
        card = Card(name=card_name)
        part.add_card(card=card, quantity=int(quantity))

    return part


# endregion


def get_deck_from_url(url: str) -> Deck:
    """
    Generates a Deck from the given URL.

    Args:
        url: The url of the web page containing a decklist.

    Returns:
        A Deck representation of the dceklist at the given URL.
    """

    parse_result = urlparse(url)
    apex_domain = get_apex_from_netloc(parse_result.netloc)

    match apex_domain:
        case "mtggoldfish":
            return get_deck_from_url_goldfish(url)
        case "mtgdecks":
            return get_deck_from_url_mtg_decks(url)
        case "mtgtop8":
            return get_deck_from_url_mtg_top8(url)
        case "scryfall":
            return get_deck_from_url_scryfall(url)
        case "tappedout":
            return get_deck_from_url_tappedout(url)
        case _:
            raise Exception("Invalid decklist URL.")


def get_deck_from_url_goldfish(url: str) -> Deck:
    """
    TODO: docstring

    main header: "Deck"
    side header: "Sideboard"
    """

    deck_id = re.findall("\d{5,10}", url)

    # raise an exception if no deck ID is found in the URL
    if len(deck_id) > 0:
        deck_id = deck_id[0]
    else:
        raise Exception("MtgGoldfish: no deck ID found.")

    goldfish_target_url = f"https://www.mtggoldfish.com/deck/arena_download/{deck_id}"
    html = get_html_document(goldfish_target_url)
    soup = BeautifulSoup(html, "html.parser")

    copy_paste_box_text = soup.find("textarea", {"class": "copy-paste-box"}).text.strip()

    deck = Deck()
    deck_parts = copy_paste_box_text.split("\n\n")
    for part in deck_parts:
        lines = part.split("\n")
        match lines[0]:
            case "Deck":
                main = deckpart_from_lines(lines[1:])
                deck.main = main
            case "Sideboard":
                side = deckpart_from_lines(lines[1:])
                deck.side = side
            case _:
                # TODO: logging
                print('MtgGoldfish: unknown decklist section found - "' + lines[0] + '". Skipping...')

    return deck


def get_deck_from_url_mtg_decks(url: str) -> Deck:
    """
    TODO: docstring

    main no header
    side header: "Sideboard"
    """

    mtg_decks_target_url = url + "/txt" if not url.endswith("/txt") else url
    decklist = get_html_document(mtg_decks_target_url)  # plaintext download

    deck = Deck()
    deck_parts = decklist.split("\n\nSideboard\n")

    main_lines = deck_parts[0].strip().split("\n")
    side_lines = deck_parts[1].strip().split("\n")

    main = deckpart_from_lines(main_lines)
    deck.main = main
    side = deckpart_from_lines(side_lines)
    deck.side = side

    return deck


def get_deck_from_url_mtg_top8(url: str) -> Deck:
    """
    TODO: docstring

    main prefix: none
    side prefix: Sideboard
    """

    deck_id = re.findall("d=\d{5,10}", url)  # NOTE: it seems like all deck IDs are 6 digits

    # raise an exception if no deck ID is found in the URL
    if len(deck_id) > 0:
        deck_id = deck_id[0]
    else:
        raise Exception("MtgTop8: no deck ID found.")

    mtg_top8_target_url = f"https://www.mtgtop8.com/mtgo?{deck_id}"
    decklist = get_html_document(mtg_top8_target_url)  # plaintext download

    deck = Deck()
    deck_parts = decklist.split("Sideboard\r\n")

    main_lines = deck_parts[0].strip().split("\r\n")
    side_lines = deck_parts[1].strip().split("\r\n")

    main = deckpart_from_lines(main_lines)
    deck.main = main
    side = deckpart_from_lines(side_lines)
    deck.side = side

    return deck


def get_deck_from_url_scryfall(url: str) -> Deck:
    """
    TODO: docstring

    main prefix:
    side prefix:
    """

    deck_id = re.findall(Constants.UUID_PATTERN, url)

    # raise an exception if no deck ID is found in the URL
    if len(deck_id) > 0:
        deck_id = deck_id[0]
    else:
        raise Exception("Scryfall: no deck ID found.")

    scryfall_target_url = f"https://api.scryfall.com/decks/{deck_id}/export/text"
    decklist = get_html_document(scryfall_target_url)  # plaintext download

    deck = Deck()
    deck_parts = decklist.strip().split("\r\n\r\n// Sideboard\r\n")

    main_lines = deck_parts[0].strip().split("\r\n")
    side_lines = deck_parts[1].strip().split("\r\n")

    main = deckpart_from_lines(main_lines)
    deck.main = main
    side = deckpart_from_lines(side_lines)
    deck.side = side

    return deck


def get_deck_from_url_tappedout(url: str) -> Deck:
    """
    TODO: docstring

    main prefix: none
    side prefix: Sideboard:
    """

    tappedout_target_url = url + "?fmt=txt" if "?fmt=txt" not in url else url
    decklist = get_html_document(tappedout_target_url)  # plaintext download

    deck = Deck()
    deck_parts = decklist.split("\n\nSideboard:\n")

    main_lines = deck_parts[0].strip().split("\n")
    side_lines = deck_parts[1].strip().split("\n")

    main = deckpart_from_lines(main_lines)
    deck.main = main
    side = deckpart_from_lines(side_lines)
    deck.side = side

    return deck



from datetime import datetime
from typing import Counter, List

from scraptrawler.deck import Deck, Format, InThe
from scraptrawler.scraper.base import BaseScraper, LoadTime
from scraptrawler.utils import Constants, ExtendedEnum, get_logger
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from strenum import StrEnum


class MtgoXPath(ExtendedEnum, StrEnum):
    """
    A class used for storing XPath related to MTGO and its DOM elements.

    Deck Parts
    ----------
    DECK : Deck part elements.
    MAIN : Main deck elements.
    SIDE : Sideboard elements.
    CARD : Card elements. Includes name and quantity.
    """

    # Event info
    DATE = "//p[@class='decklist-posted-on']"

    # Deck parts
    CARD = ".//a[@class='decklist-card-link']"
    DECK = "//div[@class='decklist-sort-group decklist-sort-type']"
    MAIN = DECK + "//div[@class='decklist-category-columns']"
    SIDE = DECK + "//ul[@class='decklist-category-list decklist-sideboard decklist-category-columns']"


class MtgoScraper(BaseScraper):
    """
    A class to represent a type of Scraper used to get data about
    Magic: the Gathering decklists and results from [MTGO](https://mtgo.com).

    Attributes
    ----------
    webdriver : webdriver.Remote
        Controls a browser by sending commands to a remote server.

    Methods
    -------
    get_decks_from_event_url(url: str, top8: bool):
        Returns a list of Decks from the given event URL.
    get_deck_from_url(url: str):
        Returns a Deck from the given deck page URL.
    """

    def __init__(self, driver: webdriver.Remote = None):
        self.logger = get_logger(self.log_filename, "MtgoScraper")
        super().__init__(driver=driver, logger=self.logger)

    ### Public Function Overrides

    def get_decks_from_event_url(self, url: str, top8: bool = False) -> List[Deck]:
        # Get page
        self._load(url)
        event_name = self._sanitized_title()
        self.logger.info(f"{url} - Processing event - {event_name}.")

        ## Get Deck parts

        # Archetype
        # TODO: get archetype with db api maybe?
        # deck_archetype = Deck.DEFAULT_ARCHETYPE

        # Format
        deck_format = Format.get(event_name.split(" ")[0])

        # Date
        date_played_element = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_element_located((By.XPATH, MtgoXPath.DATE))
        )
        date_played = datetime.strptime(
            date_played_element.text.replace("Posted on ", "").strip(), Constants.MTGO_DATE_FORMAT
        )

        # Get cards and quantities from the page
        main_elements = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_all_elements_located((By.XPATH, MtgoXPath.MAIN))
        )
        side_elements = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_all_elements_located((By.XPATH, MtgoXPath.SIDE))
        )

        decks = self._decks_from_elements(
            deck_format=deck_format, date_played=date_played, main_elements=main_elements, side_elements=side_elements
        )
        # FIXME: Is it faster to have this run against main_elements and side_elements?
        if top8:
            decks = decks[:8]

        return decks

    def get_deck_from_url(self, url: str) -> Deck:
        # Get page
        self._load(url)
        event_name = self._sanitized_title()
        self.logger.info(f"{url} - Processing deck from event - {event_name}.")

        ## Get cards and quantities from the page
        # Setup XPATH
        player = url.split("#deck_")[-1]
        player_XPATH = f"//a[@href='#deck_{player}']/ancestor::section"

        ## Get Deck parts

        # Archetype
        # TODO: get archetype with db api maybe?
        deck_archetype = Deck.DEFAULT_ARCHETYPE

        # Format
        deck_format = Format.get(event_name.split(" ")[0])

        # Date
        date_played_element = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_element_located((By.XPATH, MtgoXPath.DATE))
        )
        date_played = datetime.strptime(
            date_played_element.text.replace("Posted on ", "").strip(), Constants.MTGO_DATE_FORMAT
        )

        # Get deck from main and side elements
        main_element = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_element_located((By.XPATH, player_XPATH + MtgoXPath.MAIN))
        )
        side_element = WebDriverWait(self.driver, LoadTime.STD).until(
            EC.presence_of_element_located((By.XPATH, player_XPATH + MtgoXPath.SIDE))
        )

        # Build Deck
        deck = self._deck_from_parts(main_element, side_element)
        deck.archetype = deck_archetype
        deck.format = deck_format
        deck.date_played = date_played

        return deck

    ### Protected Function Overrides

    def _sanitized_title(self) -> str:
        try:
            title = self.driver.find_element(By.CLASS_NAME, "decklist-title").text
            date = self.driver.find_element(By.CLASS_NAME, "decklist-posted-on").text
            return f"{title} - {date.replace('Posted on ', '')}"
        except Exception as e:
            # raise (e)
            self.logger.warning(f"{self.driver.current_url} - Unable to sanitize title. Proceeding with defaults.")
            return ""

    def _decks_from_elements(
        self,
        deck_format: Format,
        date_played: datetime,
        main_elements: List[WebElement],
        side_elements: List[WebElement],
    ) -> List[Deck]:
        # Sanity check there are the same number of main decks and sideboards
        num_mains, num_sides = len(main_elements), len(side_elements)
        if num_mains != num_sides:
            self.logger.warning(
                f"""{self.driver.current_url} - There are {num_mains} main decks and {num_sides} sideboards."""
                f"""This may cause problems."""
            )

        decks: List[Deck] = []
        for i in range(len(main_elements)):
            # Get a deck from WebElements
            deck = self._deck_from_parts(main_elements[i], side_elements[i])
            # deck.archetype = deck_archetype # TODO
            deck.format = deck_format
            deck.date_played = date_played
            decks.append(deck)

        return decks

    def _deck_from_parts(self, main_element: WebElement, side_element: WebElement) -> Deck:
        main_cards = self._cards_from_element(main_element)
        side_cards = self._cards_from_element(side_element)

        # Create a Deck
        deck = Deck()
        deck.add_cards(cards=main_cards, in_the=InThe.MAIN)
        deck.add_cards(cards=side_cards, in_the=InThe.SIDE)

        return deck

    def _cards_from_element(self, deck_element: WebElement) -> List[Counter]:
        card_elements = WebDriverWait(deck_element, LoadTime.STD).until(
            EC.presence_of_all_elements_located((By.XPATH, MtgoXPath.CARD))
        )

        cards = Counter()
        for ce in card_elements:
            quantity, card = ce.text.split(" ", maxsplit=1)
            cards.update({card: int(quantity)})

        return cards
